
#include <stdio.h>

#include "windows.h"
#include "Log.h"
#include "winnt.h"
#include "heap.h"
#include "RscData.h"

static int lf_log = LF_CONSOLE;

static DWORD freeListSizes[4] = {
	0x20, 0x80, 0x200, 0xffffffff
};

static int last_error_code;

void
InternalSetLastError(int errcode)
{
	last_error_code = errcode;
}

void
HPFreeSub(HANDLE hHeap, void *pHeapRegion, DWORD size, DWORD flags)
{

}

void *
HPInit(HANDLE hHeap,void *pHeapRegion,DWORD size,DWORD flags)
{
	HEAP_HEADER pHeap;
	HEAP_ARENA  *pLast;
	HEAP_ARENA  *pArena, *pArena2;
	DWORD       FirstArena;
	FREE_LIST_HEADER *pFreeListEntry, *pFreeListArrayEnd;
	DWORD *pFreeListSize;

	/* ensure that region is aligned on a page boundary */

	pHeap = (HEAP_HEADER) pHeapRegion;

	if (!flags & 0x100) {
	   pHeap->nextHeap = 0;
	   pHeap->nextBlock = 0;
	
	   FirstArena = sizeof(HEAP_HEADER);
	
	   pHeap->signature = 0x4948;
	   pHeap->flags     = flags;
	   pHeap->dwSize      = size;

	   if (!(flags & HEAP_NO_SERIALIZE)) {
		/* InitializeCriticalSection(&pHeap->criticalSection); */	
		/* pHeap->pCriticalSection = ... */
	   }

	   pFreeListArrayEnd = (FREE_LIST_HEADER *) &pHeap->freeListArray + 
			(4 * sizeof(FREE_LIST_HEADER));

	   pFreeListEntry = pHeap->freeListArray;
	
 	   pFreeListSize  = freeListSizes; /* static array of sizes */          

	   while (pFreeListEntry < pFreeListArrayEnd ) {
		pFreeListEntry->dwSize = *pFreeListSize;
		pFreeListEntry->arena.dwSize = 0xa0000001;
		pFreeListEntry->arena.signature = 0x4846;
		pFreeListEntry->arena.prev      = 0; /* what ?*/
		pFreeListEntry->arena.next      = 0; /* what ?*/
		
		pFreeListEntry++;
		pFreeListSize++;
	    }
	
	    /* hook up hte first and last */

	    /* point at first */
	    pArena = &pHeap->freeListArray[0].arena; 

	    /* point at last */
	    pArena2 = &pHeap->freeListArray[3].arena; 

	    pArena->next = pArena2;
	    pArena2->prev = pArena;

	} else {
	
	    FirstArena = 8;

	}

	pHeap->dwSize = size;

	pLast= (HEAP_ARENA *)((int)pHeap + size - 0x10);
	
	/* verify last page committed */

	pLast->dwSize = 0xa0000000;
	pLast->signature = 0x4842;
	
	if ((!(flags & 0x400) && size > 0x1000)) {
	   size -= FirstArena;
	   size -= 0xffc;
	 
	   pArena = pHeapRegion + size + FirstArena;

	   pArena->dwSize = 0xa0000010;
	   pArena->signature = 0x4842;

	   HPFreeSub(hHeap, pArena + 0x10, 0xfdc, 0);

	} else {
	  size -= FirstArena;
	  size -= 0x10;
	}

	HPFreeSub(hHeap, pHeapRegion + FirstArena, size, 0);

	return pHeapRegion;
}

HANDLE
HeapCreate(DWORD options,DWORD isize,DWORD msize)
{
	DWORD rc;
	HANDLE hHeap;

	LOGSTR((LF_APICALL,"HeapCreate(%x,%x,%x)\n",options,isize,msize));

	if (msize && msize < isize) {

		LOGSTR((LF_APIFAIL,"HeapCreate returns DWORD 0\n"));
		InternalSetLastError(ERROR_INVALID_PARAMETER);

		return 0;
	}

	msize += 0xfff;
	msize &= 0xfffff000;

	if (msize == 0) {
		options |= HEAP_FREE_CHECKING_ENABLED;

		isize   &= 0xfffff000;
		msize    = isize +  0x10000; //0x100000; RESERVED...
	}

	//rc = (DWORD) VirtualAlloc(isize);
	rc = (DWORD) WinMalloc(msize);
	hHeap = (HANDLE) rc; 	

	/* only use these flags when calling HPInit */
	options &= (HEAP_FREE_CHECKING_ENABLED | HEAP_GENERATE_EXCEPTIONS |
		HEAP_NO_SERIALIZE);
	
	rc = (DWORD) HPInit(hHeap,(void *) hHeap,msize,options);
	 
	LOGSTR((LF_APIRET,"HeapCreate returns DWORD %p\n",rc));
	return rc;
}

// based on current running application
int
GetStartupInfoA(LPSTARTUPINFO32A lp)
{
	logstr(lf_log,"GetStartupInfoA(%p) sizeof=%x\n",
		lp,sizeof(STARTUPINFO32A));
	memset(lp,0,sizeof(STARTUPINFO32A));
	
	return 0;
}

int
HeapAlloc(int heap,int flags,int size)
{
	int rc;
	logstr(lf_log,"HeapAlloc(%x,%x,%x)\n",heap,flags,size);
	rc = (int) malloc(size);
	return rc;
}

DWORD
GetFileType32(HANDLE hFile)
{
	logstr(lf_log,"GetFileType32(%x)\n",hFile);
	return 0;	
}

HANDLE
GetStdHandle(DWORD nStdHandle)
{
	logstr(lf_log,"GetStdHandle(%x)\n",nStdHandle);
	if(nStdHandle == STD_INPUT_HANDLE)
		return INVALID_HANDLE_VALUE;
	if(nStdHandle == STD_OUTPUT_HANDLE)
		return INVALID_HANDLE_VALUE;
	if(nStdHandle == STD_ERROR_HANDLE)
		return INVALID_HANDLE_VALUE;

	return INVALID_HANDLE_VALUE;
}

UINT
GetACP(void)
{
	logstr(lf_log,"GetACP()\n");
	return 0x4e4;  // whatever this is...
}

typedef struct {
	UINT	MaxChar;
	BYTE   DefaultChar[2];
	BYTE   LeadByte[6];
} CPINFO;

typedef CPINFO *LPCPINFO;

UINT
GetCPInfo(UINT codepage,LPCPINFO lpcpinfo)
{
	logstr(lf_log,"GetCPInfo(%x,%p)\n",codepage,lpcpinfo);
	lpcpinfo->MaxChar = 1;
	return codepage;  
}

LPTSTR
GetCommandLineA(VOID)
{
	logstr(lf_log,"GetCommandLine\n");
	return "somekind of application";
}

LPTSTR
GetEnvironmentStringsW(VOID)
{
	LPTSTR penv;
	logstr(lf_log,"GetEnvironmentsStringsW\n");
	penv = (LPTSTR)  malloc(32);
	memcpy(penv,"\0\0",2);
	return penv;
}

int
FreeEnvironmentStringsW(LPSTR penv)
{
	logstr(lf_log,"FreeEnvironmentsStringsW(%p)\n",penv);
	free(penv);
	return 1;
}

int
WideCharToMultiByte(UINT codepage,DWORD dwFlags,LPCWSTR wide,int widelen,LPSTR multi,int multilen,LPCSTR defchar,LPBOOL fuseddefault)
{
	logstr(lf_log,"WideCharToMultiByte(%d,%d,%p,%d,%p,%d,%p,%p)\n",
		codepage,dwFlags,
		wide,widelen,
		multi,multilen,
		defchar,fuseddefault);
	if(fuseddefault)
		*fuseddefault = 0;

	if(multilen == 0)
		multilen = 2;
	else
		memset(multi,0,multilen);

	return multilen;
}

int
GetModuleFileNameA(HINSTANCE hInstance,LPSTR lpstr,int len)
{
	logstr(lf_log,"GetModuleFileNameA(%d,%p,%d)\n",
		hInstance,lpstr,len);

	return (int) strncpy(lpstr,"win32.exe",min(10,len));
}

int
GetModuleHandleA(HINSTANCE hInstance)
{
	return 0;
}

DWORD
MessageBoxA(HWND hWnd, LPSTR lpstr, LPSTR caption, DWORD flags)
{
	logstr(lf_log,"MessageBoxA(%x,%s,%s,%x)\n",
		hWnd,lpstr,caption,flags);

	return MessageBox(hWnd,lpstr,caption,flags);
}

int
IsBadReadPtr32(void *p,int nsize)
{
	logstr(lf_log,"IsBadReadPtr32(%p,%d)\n",p,nsize);
	return 0;
}

int
IsBadWritePtr32(void *p,int nsize)
{
	logstr(lf_log,"IsBadWritePtr32(%p,%d)\n",p,nsize);
	return 0;
}

int
GetVersion32()
{
	logstr(lf_log,"GetVersion %x\n",0x05650004);
	return 0x05650004;
}


int
LoadStringA(HINSTANCE hInstance, int wID, char *lpstr,int nCount)
{
	logstr(lf_log,"LoadString(HINSTANCE=%x,id=%d,lpstr=%p,nBytes=%d)\n",
		hInstance,wID,lpstr,nCount);
	strncpy(lpstr,"Willows NT (r)",nCount);
	return strlen(lpstr);
}


void
ShellAboutA(HWND hWnd, LPCSTR lpszCaption, LPCSTR lpszAboutText, HICON hIcon)
{
	FARPROC		shellabout;
	HINSTANCE	hInst;
	LPCSTR lpstr = lpszCaption;
	LPCSTR lptxt = lpszAboutText;

	logstr(lf_log,"ShellAboutA(HWND=%x,Caption=%p,Text=%p,HICON=%x)\n",
		hWnd,lpszCaption,lpszAboutText,hIcon);

	/* LATER: Move our icon to RscData.rc instead */
	hIcon = LoadIcon ( 0, MAKEINTRESOURCE ( IDI_WILLOWS ) );

	/* note: this will fail if we are using the REAL shell.dll */
	/*       so if we don't have shellabout, then call         */
	/*       MessageBox instead.                               */
	/* note: we need to do this, rather than a direct call to  */	
	/* 	 shell about function, it is now in a library      */
	/*       It does not have a native to binary hook 	   */
	hInst = LoadLibrary("shell");
	shellabout = GetProcAddress(hInst,"ShellAbout");

	/*								*/
 	/* Note: this will fail even if the real shell.dll is loaded 	*/
	/*	 because we don't have an OEM table for the shell       */
	/*	 functions.  A binary application can call the function */
	/*	 directly.						*/

	if(lpstr == 0)
		lpstr = "Willows TWIN/NT";
	if(lptxt == 0)
		lptxt = "Willows Software, Inc.";

	if (shellabout)
		shellabout( hWnd, lpstr,lptxt, hIcon);
	else 
		MessageBox(hWnd, lptxt, lpstr, MB_OK);
	
	if(hIcon)
		DestroyIcon ( hIcon );

}

void
Shell32End()
{}
